"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _utils = _interopRequireDefault(require("./utils"));

var _local_storage_adapter = _interopRequireDefault(require("./adapters/local_storage_adapter"));

var _google_universal_analytics_adapter = _interopRequireDefault(require("./adapters/google_universal_analytics_adapter"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class Experiment {
  constructor(options = {}) {
    this.options = options;

    _utils.default.defaults(this.options, Experiment._options);

    validate(this.options);
    this.name = this.options.name;
    this.variants = this.options.variants;
    this.user_id = this.options.user_id;
    this.variant_names = _utils.default.keys(this.variants);
    this.run();
  }

  run() {
    _utils.default.log("running...", this.options);

    const variant = this.get_stored_variant();

    if (variant) {
      // a variant was already chosen. activate it
      _utils.default.log("variant already active", {
        experiment: this.options.name,
        variant
      });

      this.activate_variant(variant);
    } else {
      this.conditionally_activate_variant();
    }
  }

  activate_variant(variant) {
    var _this$variants$varian;

    (_this$variants$varian = this.variants[variant]) === null || _this$variants$varian === void 0 ? void 0 : _this$variants$varian.activate(this);
    this.storage().set(`${this.options.name}:variant`, variant);
  } // if experiment conditions match, pick and activate a variant, track experiment start


  conditionally_activate_variant() {
    const experiment = this.options.name;
    if (!this.options.trigger()) return;

    _utils.default.log("trigger set", {
      experiment
    });

    if (!this.in_sample()) return;

    _utils.default.log("in sample", {
      experiment
    });

    const variant = this.pick_variant();

    _utils.default.log("variant picked", {
      experiment,
      variant
    });

    this.tracking().experiment_start(this, variant);
    this.activate_variant(variant);
  }

  goal_complete(goal_name, props = {}) {
    _utils.default.defaults(props, {
      unique: true
    });

    if (props.unique && this.storage().get(`${this.options.name}:${goal_name}`)) return;
    const variant = this.get_stored_variant();
    if (!variant) return;

    if (props.unique) {
      this.storage().set(`${this.options.name}:${goal_name}`, true);
    }

    _utils.default.log("experiment goal complete", {
      name: this.options.name,
      variant,
      goal: goal_name
    });

    this.tracking().goal_complete(this, variant, goal_name, props);
  }

  get_stored_variant() {
    return this.storage().get(`${this.options.name}:variant`);
  }

  pick_variant() {
    const all_variants_have_weights = _utils.default.check_weights(this.variants);

    if (all_variants_have_weights) {
      return this.pick_weighted_variant();
    }

    return this.pick_unweighted_variant();
  }

  pick_weighted_variant() {
    // Choosing a weighted variant:
    // For A, B, C with weights 1, 3, 6
    // variants = A, B, C
    // weights = 1, 3, 6
    // weights_sum = 10 (sum of weights)
    // weighted_index = 2.1 (random number between 0 and weight sum)
    // ABBBCCCCCC
    // ==^
    // Select B
    _utils.default.log("picking weighted variant", {
      experiment: this.options.name
    });

    const weights_sum = _utils.default.sum_weights(this.variants);

    let weighted_index = Math.ceil(this._random("variant") * weights_sum);

    for (const [key, value] of Object.entries(this.variants)) {
      // then we are substracting variant weight from selected number
      // and it it reaches 0 (or below) we are selecting this variant
      weighted_index -= value.weight;
      if (weighted_index <= 0) return key;
    }
  }

  pick_unweighted_variant() {
    _utils.default.log("picking unweighted variant", {
      experiment: this.options.name
    });

    const partitions = 1.0 / this.variant_names.length;
    const chosen_partition = Math.floor(this._random("variant") / partitions);
    return this.variant_names[chosen_partition];
  }

  in_sample() {
    let active = this.storage().get(`${this.options.name}:in_sample`);
    if (typeof active !== "undefined") return active;
    active = this._random("sample") <= this.options.sample;
    this.storage().set(`${this.options.name}:in_sample`, active);
    return active;
  }

  _random(salt) {
    if (!this.user_id) return _utils.default.random();
    const seed = `${this.name}.${salt}.${this.user_id}`;
    return _utils.default.random(seed);
  }

  add_goal(goal) {
    goal.add_experiment(this);
  }

  add_goals(goals) {
    for (const goal of goals) this.add_goal(goal);
  }

  storage() {
    return this.options.storage_adapter;
  }

  tracking() {
    return this.options.tracking_adapter;
  }

}

_defineProperty(Experiment, "_options", {
  name: null,
  variants: null,
  user_id: null,
  sample: 1.0,

  trigger() {
    return true;
  },

  tracking_adapter: _google_universal_analytics_adapter.default,
  storage_adapter: _local_storage_adapter.default
});

const validate = options => {
  if (options.name === null) {
    throw new Error("an experiment name must be specified");
  }

  if (options.variants === null) {
    throw new Error("variants must be provided");
  }

  if (typeof options.trigger !== "function") {
    throw new Error("trigger must be a function");
  }

  const all_variants_have_weights = _utils.default.validate_weights(options.variants);

  if (!all_variants_have_weights) {
    throw new Error("not all variants have weights");
  }
};

var _default = Experiment;
exports.default = _default;